/*
 * GWT - General Windowing Toolkit
 *
 * Copyright (C) 1998 MenTaLguY - mentalg@geocities.com
 *                    Rodolphe Ortalo - ortalo@laas.fr
 *                    Tristan Wibberley - bloater@ps.cus.umist.ac.uk
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: window.c,v $
 * Revision 1.13  2000/06/09 11:34:44  ortalo
 * Added support for drawing functions based on LibXMI.
 *
 * Revision 1.12  2000/03/13 22:02:55  ortalo
 * Committed new version of LibGWT corresponding to 0.0.2.
 * The event management API is probably more or less defined. The simplest
 * approach finally won. (I should have remembered: KISS!)
 *
 * Revision 1.11  1999/05/24 00:51:23  marcus
 * Handling of 'inp' parameter was totally bogus. Use it in a way which at
 * least makes sense.
 * Does it have a purpose at all?
 *
 * Revision 1.10  1999/05/23 23:56:43  marcus
 * Fixed broken includes.
 *
 * Revision 1.9  1999/05/20 20:34:23  ortalo
 * Several bugs corrections and small improvements.
 *
 * Revision 1.8  1999/05/13 20:01:38  ortalo
 * New version of libgwt. Now using LibGII for input management.
 * Some drawing functions implemented. Visibility management should
 * be rather okay. Still some work to do, but now, your eyes can
 * look at some drawings... not only code. ;-)
 *
 * Revision 1.7  1998/11/17 23:28:08  ortalo
 * Brought libgwt under autoconf&automake&libtool, in conformance with
 * the GNU configure and building system (and also, the rest of GGI. :-)
 * Should build and run as before (i.e.: not so much... :-(.)
 *
 * Revision 1.6  1998/11/17 15:13:28  tristan
 *
 * Changed my email address
 *
 * Revision 1.5  1998/09/02 02:21:49  tristan
 * Loadsa great stuff done. Visible/invalid regions,
 *  copy on move - fixed tt_region.c too!
 *
 * Revision 1.4  1998/08/23 18:57:15  ortalo
 * Minor editing (mainly comments).
 *
 * Revision 1.3  1998/08/16 18:40:58  ortalo
 * A new version, STILL IN VERY ALPHA STATE... (Be patient.)
 * (However, this version is a little better... :-)
 *
 * Revision 1.2  1998/08/11 22:10:11  ortalo
 * Incorporated the changes made by Tristan Wibberley. The code is
 * still in alpha development state.
 *
 * Revision 1.1  1998/07/09 18:38:39  ortalo
 * Initial integration of libgwt in the GGI repository.
 * This is experimental code, for your eyes only !
 * (Don't execute... ;-)
 *
 */

#include <ggi/internal/internal.h>
#include <ggi/xmi.h>

#include "internal.h"
#include "gwt_debug.h"

/* TODO: Threadsafety */

/*******************************************************************
 *                      STATIC FUNCTIONS                           *
 *******************************************************************/
/*
 * Window data type creation
 */
static void new_window(gwt_window_t *win)
{
  (*win) = (gwt_window_t)malloc(sizeof(gwt_window));
  gwt_zinit_window(*win);

  gwt_region_create(&(*win)->geometry);
  gwt_region_create(&(*win)->visible);
  gwt_region_create(&(*win)->potentially_visible);
  gwt_region_create(&(*win)->invalid);
  gwt_region_create(&(*win)->drawing);
  (*win)->clipping = NULL; /* No clipping initially */
#if 0
  (*win)->id = gwt_allocate_id();
  gwt_associate_id((*win)->id, *win);
#endif
  (*win)->priv = NULL;
  return;
}

/*
 * Window data type destruction
 */
static void delete_window(const gwt_window_t win)
{
  gwt_region_destroy(win->geometry);
  gwt_region_destroy(win->visible);
  gwt_region_destroy(win->potentially_visible);
  gwt_region_destroy(win->invalid);
#if 0
  gwt_free_id(win->id);
#endif
  free(win);
  return;
}

/*******************************************************************
 *                 GWT PUBLIC EXPORTED FUNCTIONS                   *
 *******************************************************************/

/*
 * Root window creation/attachment
 */
int gwtCreateRootWindow(gii_input_t inp, ggi_visual_t vis, gwt_window_t *win) {
  gwt_window_t root;
  
  /* Attach the LibXMI extension to the visual */
  if (xmiAttach(vis) < 0) {
    ggiPanic("Unable to attach XMI extension to visual\n");
  }

  new_window(win);
  root = (*win);
  root->level = 0;
  root->visual = vis;
  root->input = ggiJoinInputs(vis, inp);
  root->ptr_x = 0;
  root->ptr_y = 0;
  root->wparent = NULL;
  root->hidden = 0; /* This one is visible */
  root->implicitly_hidden = 0;
  root->focuss = NULL;

  root->origin.x = 0;
  root->origin.y = 0;
  gwt_region_init(root->geometry, 0, 0,
		  LIBGGI_MODE(vis)->visible.x,
		  LIBGGI_MODE(vis)->visible.y);
  /*
   * We must update ourselves the visibility regions.
   */
  /* For a root window: potential visibility is easy... */
  gwt_region_copy(root->potentially_visible, root->geometry);
  /* And, initially, visibility also...*/
  gwt_region_copy(root->visible, root->geometry);
  gwt_region_copy(root->drawing, root->geometry);

  gwt_region_init(root->invalid, 0, 0, 0, 0);
  gwt_invalidate(root, root->geometry);

  GWTDPRINT("GWT: Opened root window - win=%p\n", win);
	
  return 0;
}

/*
 * Root window termination
 */
/* TODO: Handle detachment from visual (I don't think there is anything TODO -- Tristan) */

/*
 * Creation of one sub window
 */
int gwtCreateWindow(gwt_window_t parent, gwt_window_t *win) {
  gii_input_t inp;

  new_window(win);
  (*win)->level = parent->level + 1;
  /* This is necessary if memory is not clean even if the
     real new parent is initialized later in gwtReparent */
  (*win)->wparent = NULL;
  (*win)->hidden = 1;
  (*win)->implicitly_hidden = parent->implicitly_hidden
    || parent->hidden;
  (*win)->visual = parent->visual;
  (*win)->focuss = NULL;

  /* Set Geometry to that of its parent initially... ? */
  gwt_region_copy((*win)->geometry, parent->geometry);
  gwt_region_init((*win)->visible, 0, 0, 0, 0);
  gwt_region_init((*win)->potentially_visible, 0, 0, 0, 0);
  gwt_region_init((*win)->invalid, 0, 0, 0, 0);
  gwt_region_init((*win)->drawing, 0, 0, 0, 0);

  /* Creates an empty input queue */
  inp = giiOpen("input-null", NULL);
  (*win)->input = inp;
  (*win)->ptr_x = parent->ptr_x;
  (*win)->ptr_y = parent->ptr_y;

  GWTDPRINT("GWT: About to insert child window\n");
  /* This should put the child at the top and set up its visibility */
  gwtReparent((*win), parent);
  GWTDPRINT("GWT: Inserted child window\n");
  GWTDPRINT("GWT: Opened sub window - win=%p\n", (*win));

  gwt_invalidate((*win), (*win)->geometry);

  return 0;
}

/*
 * Destruction of a window
 */
static void _gwt_destroy1(const gwt_window_t win) {
  gwt_window_t child;
  /* TODO: destroy children, invalidate uncovered regions */
  /* Destroy starting from back */
  child = GWT_Z_BOTTOM_CHILD(win);
  while (child != NULL)
    { /* do this first because gwt_zextract will NULL it */
      const gwt_window_t next_child = GWT_Z_UP(child); 
      gwt_zextract(child); /* Remove from zorder */
      _gwt_destroy1(child); /* Recursive destruction */
      child = next_child;
    }
  GWTDPRINT("GWT: Closing window - win=%p\n", win);
  delete_window(win);
  return; 
}

int gwtDestroy(gwt_window_t win)
{
  if (win->level > 0)
    {
      gwtHide(win);
      gwt_zextract(win);
    }
  _gwt_destroy1(win); /* NB: Don't recurse within a lock... */
  GWTDPRINT("GWT: Closed window - win=%p\n", win); 
  return 0;
}

/*
 * Window visual access
 */
int gwtGetVisual(gwt_window_t win, ggi_visual_t *vis)
{
  (*vis) = win->visual;
  return 0;
}

/*******************************************************************
 *                  GWT PRIVATE EXPORTED FUNCTIONS                 *
 *******************************************************************/
/*
 * Global lock/unlock functions for the library
 */
 /* TODO: locking */
